import pandas as pd 
import numpy as np 
from numpy.linalg import norm
from sklearn.metrics import ndcg_score, roc_auc_score, roc_curve
import surprise

def simil_cosine(a,b):
    return np.dot(a, b)/(norm(a)*norm(b))

def pearson_corr(a,b):
    return np.corrcoef(a, b)[0,1]

# compute RMSE
def rmse(pred, real):
    return np.sqrt(((pred - real) ** 2).mean())

# compute MAE
def mae(pred, real):
    return np.absolute(np.subtract(real, pred)).mean()

# check if input is pandas DataFrame
def isDataFrame(x):
    return isinstance(x, pd.DataFrame)

# convert predictions from pandas DataFrame to surprise.Prediction
def df2surprise(df):
    return [surprise.Prediction(*i) for i in df.itertuples(index=False)]

# Wrapper for accuracy metrics
def prediction_metrics(prediction, excl_impossible=False):

    if isDataFrame(prediction):
        prediction = df2surprise(prediction)

    pred = []
    real = [] 

    res = {}
    if excl_impossible:
        if all(isinstance(i, surprise.Prediction) for i in pred):

            # filter impossible prediction (np.array([True, False, True, ...])) 
            for i in prediction:
                if not i.details["was_impossible"]:
                    pred.append(i.est)
                    real.append(i.r_ui) 
            print(f'Excluded {len(prediction) - len(pred)} ({len(prediction)}) samples. {len(pred)} remaining ...')
        else:
            raise Exception(f"Argument 'prediction' not of type -> List[surprise.Prediction]")
    else:

        for i in prediction:
            pred.append(i.est)
            real.append(i.r_ui)

    # compute accuracy metrics
    res['RMSE'] = [rmse(np.array(pred), np.array(real))]
    res['MAE'] = [mae(np.array(pred), np.array(real))]
    
    # return pd.DataFrame
    return pd.DataFrame(res, index=['value']).T


def classification_metrics(prediction, threshold, topn=False, excl_impossible=False):

    if isDataFrame(prediction):
        prediction = df2surprise(prediction)

    list_metrics = ['Recall', 'Precision', 'F1']

    if topn:
        print('Warning: TopN classification not recommended to use ...')

        if excl_impossible:
            prediction = [i for i in prediction if not i.details["was_impossible"]]

        df_pred_sort = {}
        df_real_sort = {}
        df_pred = pd.DataFrame(prediction)

        # sort user preferences by rating
        for user in df_pred['uid'].unique():
            # predicted values
            plist = list(df_pred.loc[df_pred['uid']==user,:].sort_values('est', ascending=False)['iid'])
            df_pred_sort[user] = [plist]
            
            # real values
            rlist = list(df_pred.loc[df_pred['uid']==user,:].sort_values('r_ui', ascending=False)['iid'])
            df_real_sort[user] = [rlist]

        df_pred_sort = pd.DataFrame(df_pred_sort, index=['item']).T
        df_real_sort = pd.DataFrame(df_real_sort, index=['item']).T
        
        # check for consistency in recommended items
        if (len(set([len(i[:threshold]) for i in df_pred_sort["item"]])) > 2):
            print('Warning: Missing items in TopN predictions may lead to imprecise metrics ...')

        list_tp = []
        list_fp = []
        list_fn = []

        # compute overlap recommendations and true user preferences
        for pred, real in zip(df_pred_sort["item"], df_real_sort["item"]):
            common = len(np.intersect1d(pred[:threshold], real[:threshold]))
            list_tp.append(common)
            list_fp.append(len(pred[:threshold]) - common)
            list_fn.append(len(real[:threshold]) - common)
        TP = np.sum(list_tp)
        FP = np.sum(list_fp)
        FN = np.sum(list_fn)

        list_metrics = [f'Recall@{threshold}', f'Precision@{threshold}', f'F1@{threshold}']
    
    else:
        pred = []
        real = [] 

        if excl_impossible:
            if all(isinstance(i, surprise.Prediction) for i in pred):

                # filter impossible prediction (np.array([True, False, True, ...])) 
                for i in prediction:
                    if not i.details["was_impossible"]:
                        pred.append(i.est)
                        real.append(i.r_ui) 
                print(f'Excluded {len(prediction) - len(pred)} ({len(prediction)}) samples. {len(pred)} remaining ...')
            else:
                raise Exception(f"Argument 'prediction' not of type -> List[surprise.Prediction]")
                
        else:
            for i in prediction:
                pred.append(i.est)
                real.append(i.r_ui)

        pred = np.array(pred)
        real = np.array(real)
        
        # compute confusion matrix
        TP = np.sum((pred >= threshold) & (real >= threshold))
        FP = np.sum((pred >= threshold) & (real < threshold))
        FN = np.sum((pred < threshold) & (real >= threshold))
        TN = np.sum((pred < threshold) & (real < threshold))


    # compute Recall, Precision, F1
    recall    = TP / (TP+FN)
    precision = TP / (TP+FP)
    f1        = (2*precision*recall) / (precision+recall)

    df = pd.DataFrame([recall, precision, f1], index=list_metrics, columns=['value'])

    return df

def ndcg(prediction, real, topn):
    res = []
    for rl, pred in zip(real, prediction):
        # overlap between predictions and real preferences
        common = set(rl).intersection(set(pred))
        # extract ranking of common items
        rl_rank = [rl.index(x) for x in common]
        pred_rank = [pred.index(x) for x in common]

        # compute NDCG
        if len(rl_rank) == 1: 
            if rl_rank==pred_rank: res.append(1.0)
            else: res.append(0.0)
        else:
            res.append(ndcg_score([rl_rank[:topn]], [pred_rank[:topn]]))
    return np.mean(res)

# Removed - Possible to compute in theory but practical value is very limited for RecSys
# def auc(prediction, real, threshold, plot=True, excl_impossible=False, details=''):

#     if excl_impossible:
#         # check shapes
#         if len(details) != len(prediction):
#             raise Exception('Shapes do not match: prediction={}, details={}'.format(len(prediction), len(details)))
#         else:
#             # filter impossible prediction (np.array([True, False, True, ...])) 
#             arr_filt = np.array([False if i['was_impossible'] else True for i in details])
#             print('Excluded {} ({}) samples. {} remaining ...'.format(len(arr_filt) - np.sum(arr_filt),len(arr_filt), np.sum(arr_filt)))

#             # filter predictions and real values
#             prediction_filt, real_filt = np.array(prediction)[arr_filt], np.array(real)[arr_filt]
#             pred = [1 if i >= threshold else 0 for i in prediction_filt]
#             rl = [1 if i >= threshold else 0 for i in real_filt]

#     else:
#         pred = [1 if i >= threshold else 0 for i in prediction]
#         rl = [1 if i >= threshold else 0 for i in real]
    
#     score = roc_auc_score(rl, pred)

#     if plot:
#         fpr, tpr, _ = roc_curve(rl, pred)
#         plt.figure()
#         plt.plot(fpr, tpr, color='darkorange',
#                 lw=2, label='ROC curve (area = {})'.format(round(score,3)))
#         plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
#         plt.xlim([0.0, 1.0])
#         plt.ylim([0.0, 1.0])
#         plt.xlabel('False Positive Rate')
#         plt.ylabel('True Positive Rate')
#         plt.legend(loc="lower right")
#         plt.show()
    
#     # Return AUC score
#     return score


def ranking_metrics(prediction, threshold, excl_impossible=False):

    if isDataFrame(prediction):
        prediction = df2surprise(prediction)

    if excl_impossible:
        prediction = [i for i in prediction if not i.details["was_impossible"]]

    df_pred_sort = {}
    df_real_sort = {}
    df_pred = pd.DataFrame(prediction)

    # sort user preferences by rating
    for user in df_pred["uid"].unique():
        # predicted values
        plist = list(df_pred.loc[df_pred["uid"]==user,:].sort_values("est", ascending=False)["iid"])
        df_pred_sort[user] = [plist]
        
        # real values
        rlist = list(df_pred.loc[df_pred["uid"]==user,:].sort_values("r_ui", ascending=False)["iid"])
        df_real_sort[user] = [rlist]

    df_pred_sort = pd.DataFrame(df_pred_sort, index=["iid"]).T
    df_real_sort = pd.DataFrame(df_real_sort, index=["iid"]).T

    res = {}

    res[f'NDCG@{threshold}'] = ndcg(df_real_sort["iid"], df_pred_sort["iid"], threshold)

    return pd.DataFrame(res, index=['value']).T

def evaluate(prediction, topn, rating_cutoff, excl_impossible=False):
    if isDataFrame(prediction):
        prediction = df2surprise(prediction)
    return prediction_metrics(prediction, excl_impossible=excl_impossible).append(classification_metrics(prediction, rating_cutoff, excl_impossible=excl_impossible)).append(ranking_metrics(prediction, topn, excl_impossible=excl_impossible))




from collections import defaultdict
import pandas as pd

# Adapt the get_top_n function from the surprise package
def get_top_n(predictions, n=15, only_new=False, train='', user_col='user', item_col='item'):
    """Return the top-N recommendation for each user from a set of predictions.
    Args:
        predictions(list of Prediction objects): The list of predictions, as
            returned by the test method of an algorithm.
        n(int): The number of recommendation to output for each user. Default
            is 15.
    Returns:
    A dict where keys are user (raw) ids and values are lists of tuples:
        [(raw item id, rating estimation), ...] of size n.
    """
    top_n = defaultdict(list)

    # recommend only new items
    if only_new:
        if isinstance(train, pd.DataFrame):
            tagg = train.groupby(user_col).agg({item_col:lambda x: list(x)}).reset_index()
            tdict = dict(zip(tagg.user, tagg.item))
        else:
            raise Exception('ArgumentError: Train is not pd.DataFrame.')

        # Map the predictions to each user.
        for uid, iid, true_r, est, _ in predictions:
            if (uid in tdict.keys()) and (iid not in tdict[uid]):
                top_n[uid].append((iid, est))

    else:
        # Map the predictions to each user.
        for uid, iid, true_r, est, _ in predictions:
            top_n[uid].append((iid, est))

    # Then sort the predictions for each user and retrieve the k highest ones.
    for uid, user_ratings in top_n.items():
        user_ratings.sort(key=lambda x: x[1], reverse=True)
        top_n[uid] = user_ratings[:n]

    # topn to df
    df_topn = pd.DataFrame([(key, [v[0] for v in val]) for key, val in top_n.items()], columns=['user', 'item'])
    return df_topn

def predict_user_topn(model, train, user, topk=15, item_col='item'):
    """
    Returns topk items for specified user.
    Return type: list

    Args[model, train, user, topk, item_col]
    model:      fitted model (surprise package)
    train:      train set used to fit model
    user:       user id
    topk:       topk items to return
    item_col:   column containing item ids 
    """

    # intermediate result dict 
    res = {item_col:[], 'pred':[], 'detail':[]}
    
    # iter through all items contained in train set
    for item in set(train[item_col]) :
        uid, iid, true_r, est, detail = model.predict(user,item)
        if detail['was_impossible']: continue
        # save to result dict
        res[item_col].append(item)
        res['pred'].append(est)
        res['detail'].append(detail)

    return list(pd.DataFrame(res).sort_values('pred', ascending=False)[:topk][item_col])


import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
import pandas as pd
import surprise

class ContentBased:

    def __init__(self, NN):
        self.NN = NN
        self.fitted = {"content":False, "ratings":False}
        
    def fit(self, content_data):

        self.items = content_data.index.values
        self.item_dim = len(self.items)
        # check for duplicate items
        assert (len(self.items) == len(set(self.items))), "Duplicate items in content data!"

        # compute similarity
        self.matrix = cosine_similarity(content_data.values)
        np.fill_diagonal(self.matrix, 0)
        
        self.matrixNN = self.matrix.copy()

        # filter similarity matrix for NN nearest neighbors (constraint: non-negative similarity)
        for i in range(self.item_dim):
            crit_val = max(-np.sort(-self.matrix[i])[self.NN-1], 0)
            self.matrixNN[i][self.matrixNN[i] < crit_val] = 0.0
        
        self.fitted["content"] = True

    # helper -> transform surprise.trainset.Trainset to pd.DataFrame
    def _trainset2list(self, trainset):
        return pd.DataFrame([(trainset.to_raw_uid(u), trainset.to_raw_iid(i), r) for (u, i, r) in trainset.all_ratings()], columns=["user", "item", "rating"])

    def fit_ratings(self, df):

        if not self.fitted["content"]:
            raise Exception("Fit model on content data!")

        if isinstance(df, surprise.trainset.Trainset):
            df = self._trainset2list(df)
        
        # fix unknown items
        unknown_items = list(set(df["item"]) - set(self.items))
        if len(unknown_items) > 0:
            print(f"Warning {len(unknown_items)} items are not included in content data: {unknown_items}")
        df = df[df["item"].isin(self.items)].reset_index(drop=True)

        # store user data
        self.users = np.unique(df["user"])
        self.user_dim = len(self.users)

        # fix missing items
        missing_items = list(set(self.items) - set(df["item"]))
        if len(missing_items) > 0: 
            fix_df = pd.DataFrame([{"user":np.nan, "item":i, "rating":np.nan} for i in missing_items])
            df = df.append(fix_df).reset_index(drop=True)

        # pivot 
        df_pivot = df.pivot_table(index='user', values='rating', columns='item', dropna=False).reindex(self.users)

        # row-wise (user) average
        self.user_avg = np.array(np.mean(df_pivot, axis=1))
        self.global_mean = np.mean(self.user_avg)

        # center ratings
        df_pivot = df_pivot.sub(self.user_avg, axis=0).fillna(0)

        # predict ratings for each item 
        denom = self.matrixNN.sum(axis=0) # column sums
        self.prediction = (np.matmul(df_pivot.values, self.matrixNN) / denom) + self.user_avg[:,np.newaxis]

        # replace NA values with mean
        # prediction[np.isnan(prediction)] = self.global_mean

        self.fitted["ratings"] = True
    
    # get predicted value for user-item combination
    def predict(self, user, item, r_ui=None):
        details = {"was_impossible":False}

        # check whether user and item are unknown -> default = global average rating
        if self.knows_user(user) & self.knows_item(item):

            # convert user & item in internal ids
            iid = np.where(self.items == item)[0].item()
            uid = np.where(self.users == user)[0].item()

            # inference prediction
            est = self.prediction[uid, iid]
            
            if np.isnan(est): 
                est = self.global_mean
                details["was_impossible"] = True
            return surprise.Prediction(user, item, r_ui, est, details)
        
        else:
            details["was_impossible"] = True
            details["reason"] = "User or item unknown"
            return surprise.Prediction(user, item, r_ui, self.global_mean, details)

    # predict entire testset
    def test(self, testset):
        if not self.fitted["ratings"]:
            raise Exception("Fit model on ratings data!")
        return [self.predict(user=u,item=i,r_ui=r) for (u,i,r) in testset]

    def knows_user(self, user):
        return user in self.users   

    def knows_item(self, item):
        return item in self.items        

    # get topn most similar items 
    def get_most_similar(self, item, topn=5):

        # get iid
        if self.knows_item(item):
            iid = np.where(self.items == item)[0].item()
        else:
            raise Exception(f"Item {item} unknown ...")
        
        list_iids = (-self.matrix[iid]).argsort()[:topn]
        return self.items[list_iids]

    def get_similarities(self):
        print('Cosine similarities shape: ({}, {}) items x items'.format(self.item_dim, self.item_dim))
        return self.matrix


from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='ieseg_recsys',
    version='0.23',
    license='MIT',
    author="Philipp Borchert",
    author_email='p.borchert@ieseg.fr',
    packages=find_packages(),
    description = 'Recommendation Systems - IESEG School of Management',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/pnborchert',
    keywords='Recommender Systems IESEG',
    install_requires=[
          'scikit-learn',
          'scikit-surprise',
          'pandas',
          'numpy',
      ],

)